# Ringing::PlaceNotation - manipulate place notation

package Ringing::PlaceNotation;

use 5.6.0;
use strict;
use warnings;

our $VERSION = '0.01';

=head1 NAME

Ringing::PlaceNotation - manipulate place notation

=head1 SYNOPSIS

  use Ringing::PlaceNotation;

  $pn = new Ringing::PlaceNotation(6, '&-3-4-25-3-4-5,2');
  print $pn->format(symmetry => 1, cross => 'x', external => 1);

=head1 DESCRIPTION

This module provides a class to represent a piece of place notation.
It can parse input in a flexible format and provide output in a choice
of different formats.

We also provide some separate functions for performing various
operations on place notation strings and bell numbers.

=cut

require Exporter;
our @ISA = qw(Exporter);
our @EXPORT_FUNCS = qw(bell_to_string string_to_bell expand_pn);
our @EXPORT_RES = qw($any_bell $odd_bell $even_bell
		     $simple_pn $compound_pn);
our @EXPORT_OK = (@EXPORT_FUNCS, @EXPORT_RES);
our %EXPORT_TAGS = ( all => \@EXPORT_OK,
                     funcs => \@EXPORT_FUNCS,
                     res => \@EXPORT_RES );

=head2 EXPORT

None by default.

All the functions documented below are available for export; the
tag B<:funcs> will export them all.

There are also some precompiled regular expressions which may be
exported, as described below; they can all be exported by using
the tag B<:res>.

The tag B<:all> exports both the functions and the regular expressions.

=cut

our $bells = '1234567890ETABCDFGHJKLMNPQRSUVWYZ';

our $any_bell = qr/[$bells]|{\d+}/o;
our $odd_bell = qr/[13579EACFHKMPRUWZ]|{\d*[13579]}/;
our $even_bell = qr/[24680TBDGJLNQSVY]|{\d*[24680]}/;
our $simple_pn = qr/([-Xx]|$any_bell)+(\.([-Xx]|$any_bell)+)*/o;
our $compound_pn = qr/&? \s* $simple_pn (\s* , \s* &? \s* $simple_pn)* /ox;

=head1 PLACE NOTATION SYNTAX

A single bell is usually represented by one character.  The characters
representing bells from 1 to 33 are, as used, by the Central Council,

  1234567890ETABCDFGHJKLMNPQRSUVWYZ.

This package additionally supports an extended notation for bells above 33.
Any bell may be represented by a decimal number in curly braces, for example
B<{41}> or B<{130}>.

Place notation consists of a sequence of B<changes>.  A change is a
sequence of bells, each representing a place made in the change.  The
places should be arranged in ascending order; this is not checked but
often assumed.  A `cross' change in which no places are made may be
represented by and of the characters B<X>, B<x> or B<->.  Changes are
separated by dots, but dots either side of cross changes are optional.
External places may be omitted.  White space is permitted anywhere in
the place notation and will be ignored.  A piece of place notation
matching this description is called I<simple>.

Compound place notation is a series of pieces of place notation separated
by commas.  Each piece may start with an ampersand (B<&>), in which case
it represents a symmetrical sequence of changes:  for example, the
two strings B<&-2-3-4,5> and B<-2-3-4-3-2-5> are equivalent.

=head1 REGULAR EXPRESSIONS

This module defines the following regular expressions:

=over

=item B<$any_bell>

This regular expression matches any bell: either one of the characters
used by the Central Council to refer to a bell (for bells up to 33) or
a decimal number enclosed in braces (such as B<{41}>) for arbitrary
bell numbers.

=item B<$odd_bell>

This regular expression matches any odd bell.

=item B<$even_bell>

This regular expression matches any even bell.

=item B<$simple_pn>

This regular expression matches a piece of simple place notation.

=item B<$compound_pn>

This regular expression matches a piece of compound place notation.

=back

=head1 FUNCTIONS

The following functions are defined by this module.

=cut

sub bell_to_string {
  my $b = shift;
  return ($b <= 33) ? substr($bells, $b, 1) : "{$b}";
}

=head2 bell_to_string()

  $s = bell_to_string($b);

This function takes a number (at least 1) and returns the string
representing that bell; for example 12 becomes 'T'.

=cut

sub string_to_bell {
  my $s = shift;
  if($s =~ /^{(\d+)}$/) { return $1; }
  return index($bells, $s);
}

=head2 string_to_bell()
  
  $b = string_to_bell($s);

This function takes a string representing one bell and returns the
number of that bell.  If the string is not recognised, it returns 0.

=cut

sub maximise_pn {
  my ($b, $pn) = @_;
  $pn =~ s/\s+//g;
  my $sym = ($pn =~ s/^&//);
  $pn =~ tr/-Xx/X/;
  $pn =~ s/X/.X./g;
  $pn = ".$pn.";
  $pn =~ tr/.//s;
  $pn =~ s/\.(?=$even_bell)/.1/g;
  my $x = bell_to_string($b-1);
  if($b & 1) {
    $pn =~ s/($even_bell)\./$1$x./g;
  } else {
    $pn =~ s/($odd_bell)\./$1$x./g;
  }
  $pn =~ s/^\.//;
  $pn =~ s/\.$//;
  return $sym ? "&$pn" : $pn;
}

# Assumes we have one piece of maximal place notation
sub format_pn {
  my ($b, $pn, $options) = @_;
  $pn = ".$pn.";
  do {
    my $x = bell_to_string($b-1);
    $x =~ s/{/\\{/g;
    if($b & 1) {
      $pn =~ s/($even_bell)$x(?=\.)/$1/g;
    } else {
      $pn =~ s/($odd_bell)$x(?=\.)/$1/g;
    };
    $pn =~ s/\.1(?=$even_bell)/./g;
  } unless $options->{external};		 
  $pn =~ s/\.X\./X/g unless $options->{dots};
  do {
    $pn =~ tr/X/x/ if $options->{cross} eq 'x';
    $pn =~ tr/X/-/ if $options->{cross} eq '-';
  } if $options->{cross};
  $pn =~ s/^\.//;
  $pn =~ s/\.$//;
  return $pn;
}

sub expand_once {
  my $pn = shift;
  do {
    my @changes = split /\./, $pn;
    my $hl = pop @changes;
    $pn = join '.', @changes, $hl, reverse @changes;
  } if $pn =~ s/^&//;
  return $pn;
}

sub expand_pn {
  my ($b, $pn) = @_;
  return join '.', map(expand_once(maximise_pn($b, $_)), split(/,/, $pn));
}

=head2 expand_pn()

  my $s = expand(pn($bells, $pn);

This function takes a piece of I<compound> place notation (see above) and
puts it into maximal form - that is, with dots between all changes including
cross changes, with external places, and with cross changes represented by 
B<X>.  For example, B<&-2-3,4> becomes B<X.2.X.3.X.2.X.4>.

=cut

sub is_rotation {
  my ($x, $y) = @_;
  return undef unless length($x) == length($y);
  my $i = index("$y.$y", $x);
  return ($i < 0) ? undef : substr($y, 0, $i) =~ tr/././;
}

sub backwards {
  return join '.', reverse split(/\./, $_[0]);
}

sub is_symmetric {
  my $pn = shift;
  my $l = ($pn =~ tr/././) + 1;
  return undef if $l & 1;
  my $i = is_rotation($pn, backwards($pn));
  return undef unless defined $i;
  return ($i & 1) ? (($l - $i + 1) >> 1) : undef;
}

sub split_at {
  my ($pn, $i) = @_;
  return ('', $pn) if $i == 0;
  my $p = 0;
  for(;$p >= 0 && $i > 0; $i--) {
    $p = index($pn, '.', $p+1);
  }
  return ($p < 0) ? ($pn, '') 
    : (substr($pn, 0, $p), substr($pn, $p+1));
}

# ------------ class PlaceNotation starts here --------------

=head1 THE Ringing::PlaceNotation CLASS

The B<Ringing::PlaceNotation> class represents a piece of place notation.
It encapsulates the preceding functions in a single class definition.

=head2 METHODS

The class defines the following methods.

=cut

sub new {
  my $proto = shift;
  my $class = ref($proto) || $proto;
  my $self = {};
  $self->{b} = shift;
  $self->{pn} = shift;
  bless $self, $class;
  $self->canonicalise;
  return $self;
}

=head3 Constructor

  $pn = new PlaceNotation($bells, $pn_string);

The constructor takes a number of bells and a string of compound place
notation (see above).

=cut

sub canonicalise {
  my $self = shift;
  return unless $self->{pn};
  $self->{pn} = expand_pn($self->{b}, $self->{pn});
}

sub maximal {
  my $self = shift;
  return $self->{pn};
}

=head3 maximal()

  $s = $pn->maximal();

This function returns the place notation in maximal format.  This is how
it is stored internally, so there is no overhead as there is for B<format()>.

=cut

sub format {
  my $self = shift;
  $self->{pn} or return undef;
  my %options = @_;
  if($options{symmetry}) {
    my $i = is_symmetric($self->{pn});
    if(defined $i) {
      my ($a, $b) = split_at($self->{pn}, $i);
      (undef, $b) = split_at($b, $i - 1);
      ($b, undef) = split_at($b, (($b =~ tr/././) + 2) >> 1);
      my $as = ($a =~ /\./) ? '&' : '';
      my $bs = ($b =~ /\./) ? '&' : '';
      $a = format_pn($self->{b}, $a, \%options);
      $b = format_pn($self->{b}, $b, \%options);
      return "$as$a,$bs$b";
    } else {
      return format_pn($self->{b}, $self->{pn}, \%options);
    }
  } else {
    return format_pn($self->{b}, $self->{pn}, \%options);
  }
}

=head3 format()

  $s = $pn->format(%options);

This function returns a string containing the place notation, formatted
according to the options given in a hash.  The following entries in the
hash are recognised:

=over

=item dots

If true, put dots between all changes (including cross changes).

=item external

If true, include external places; otherwise, remove them.

=item cross

This hash entry may be B<X>, B<x> or B<->; it controls which character
is used to represent a cross change.  The default is B<X>.

=item symmetry

If this is true, the place notation is checked for symmetry.  If it is
symmetrical, the result is a string of compound place notation consisting
of two pieces, each symmetrical.  If the place notation is not symmetrical,
this option has no effect.

=back

For example,

  my $pn = new Ringing::PlaceNotation(6, '3-4-25-3-4-5-4-3-25-4-3-2');
  print $pn->format(symmetry => 1, cross => 'x', external => 1);

will print B<&36x14x1256x36x14x56,12>.

=cut

1;
__END__


=head1 AUTHOR

Martin Bright, E<lt>martin@boojum.org.ukE<gt>

=head1 COPYRIGHT AND LICENCE

Copyright (C) 2004 by Martin Bright.  All rights reserved.

This library is free software; you can redistribute it and/or modify
it under the same terms as Perl itself, either Perl version 5.8.2 or,
at your option, any later version of Perl 5 you may have available.


=cut
